var ejs = require('ejs');
var fs = require('fs');
var events = require("events");
var util   = require('util');
var _ = require('underscore');
var dutil = require('node-xmpp-bosh').dutil;

var getTemplate = function(path) {
  return fs.readFileSync(__dirname + '/templates/' + path, 'utf8');  
};

var queryElement = function(c){
    return c.name === 'query';
};

var deserialize = function(template, locals) {
    return dutil.xml_parse(render(template, locals).replace(/\n|\t|^\s*|\s(?=\s)|\s*$/g,""));
};

var render = function(template, locals) {
    return ejs.render(getTemplate(template), {
        locals: locals
    });
};

var State = function(state) {
}

State.prototype = {
	shouldRestartStream: function() {
		return false;
	},
    _createStanzas: function() {
        return [this._deserialize()];
    },
    popStanza: function() {
        if (this.done()) return;
        if (!this._stanzas) {
            this._stanzas = this._createStanzas();
        }
        return this._stanzas.pop();
    },
    done: function() {
        return false;
    },
    _deserialize: function() {
        return deserialize(this._template(), this._locals());
    },
    _locals: function() {},

    changeUsing: function(response) {
        var nextState = this._nextState(response);
        if (nextState) {
            this._appendStateData(response);
            return new module.exports[nextState](this._stateData);
        }
    }
};

var makeState = function(name, proto, ctor) {
    var newState = ctor || function(stateData){this._stateData = stateData || {};}
    util.inherits(newState, State);
    newState.prototype.__defineGetter__("name", function(){
             return name;
         });
    _.extend(newState.prototype, proto);
    module.exports[name] = newState;
}

makeState('PreAuth',{
    _template: function() {
        return 'auth';
    },
	_isNode: function(name) {
		return function(node) {
			return node.name === name;
		}
	},
	_auth_stanza: function() {
	    return _(this._stateData['input']).find(this._isNode("auth"));
	},
	_jid: function() {
	    return _(this._auth_stanza().children).find(this._isNode("username")).getText();
	},
	_password: function() {
	    return _(this._auth_stanza().children).find(this._isNode("password")).getText();
	},
	_token: function() {
	  return _(this._auth_stanza().children).find(this._isNode("meta-token")).getText();  
	},
	_user_id: function() {
        var index = this._jid().indexOf('@');
        return this._jid().substring(0, index);
	},
    _locals: function() {
        return {token: new Buffer(this._jid()+"\u0000"+this._user_id()+"\u0000"+this._password()+"\u0000"+this._token(), 'utf8').toString('base64')}
    },
    _appendStateData: function(response) {
        this._stateData['auth'] = response;
        this._stateData['jid'] = this._jid();
    },
    changeUsing: function(response) {
        switch(response.name) {
			case 'failure':
			    this._appendStateData(response);
			    return new module.exports['AuthFail'](this._stateData);
			case 'success':
			    this._appendStateData(response);
			    return makeComposite(['AuthSuccess', 'PreRestart'], this._stateData);
		}	
	} 
});

makeState('PreRestart',{
    shouldRestartStream: function(){
        return true;
    },
    _isBindPacket: function(node) {
        return node.name === 'bind' && node.attrs['xmlns'] === 'urn:ietf:params:xml:ns:xmpp-bind';
    },
    _appendStateData: function(response){
        this._stateData['stream_features'] = response;
    },
    _nextState: function(response) {
        if (response.name === 'stream:features')
        return _(response.children).any(this._isBindPacket) ? 'PreBind' : 'PreRoster';
    },
    popStanza: function() {
        return
    }
});

makeState('PreBind', {
    _isResourceNode: function(node){
        return node.name === 'resource';
    },
    _template: function(){
        return _(this._stateData['input']).any(this._isResourceNode) ? 'bind_with_resource' : 'bind_without_resource';
    },
    _id:  'lb_bind_1',
    _locals: function(){
        return {
            id: this._id, 
            resource: _(this._stateData['input']).chain().select(this._isResourceNode).invoke('getText').join('').value()
    }},

	_isBindResponse: function(response){
		return response.name === 'iq' && response.attrs['id'] === this._id;
	},
	
    _successful: function(response){
        return response.attrs['type'] === 'result';
    },
    _hasSessionSupport: function() {
        return _(this._stateData['stream_features'].children).any(function(node){
            return node.name === 'session' && node.attrs['xmlns'] === 'urn:ietf:params:xml:ns:xmpp-session';
        })
    },
    _appendStateData: function(response) {
        this._stateData['bind'] = response;
        this._stateData['jid'] = _(response.children).chain().
            pluck('children').
            flatten().
            find(function(node){
                return node.name === 'jid'
            }).value().
            getText()
    },
    changeUsing: function(response) {
		if (!this._isBindResponse(response)) return;
		this._appendStateData(response);
        return !this._successful(response) ? 'BindError' : 
        	(this._hasSessionSupport() ? makeComposite(['BindComplete', 'PreSession'], this._stateData) : makeComposite(['BindComplete', 'PreRoster'], this._stateData));
    }
});

makeState('PreSession', {
	_template: function(){
        return 'session';
    },

    _locals: function(){
        return {to: this._stateData['to'], id: this._id};
	},
	
	_id: 'lb_sess_1',
	
	_isSession: function(response) {
		return response.name === 'iq' && response.attrs['id'] === this._id;
	},
	
	_appendStateData: function(response) {
	    this._stateData['session'] = response;
	},
    changeUsing: function(response) {
		if (!this._isSession(response)) return;
		this._appendStateData(response);
		return makeComposite(['PreRoster', 'PreDomainItemsDisco'], this._stateData);
    }
});

makeState('PreRoster', {
    _template: function() {
        return 'roster';
    },
    _locals: function() {
        return {jid: this._stateData['jid'], id: this._id};
    },
    _id: 'lb_roster_get_1',
    _isRoster: function(response){
        return response.name === 'iq' && response.attrs['id'] === this._id;
    },
    _appendStateData: function(response) {
        this._stateData['roster'] = response;
    },
    _isVCardRequest: function(node) {
        return node.name === 'fetch-vcards' && node.attrs['xmlns'] === 'http://talk.to/x';
    },
    changeUsing: function(response) {
        if (!this._isRoster(response)) return
		this._appendStateData(response);
		return (_(this._stateData['input']).any(this._isVCardRequest)) ?
		    makeComposite(['RosterFetched', 'PreVCard'], this._stateData) :
		    new module.exports['RosterFetched'](this._stateData);
    }
});

makeState('PreVCard', {
    _createStanzas: function() {
        var jids = this._rosterJids();
        return _(jids).chain().
        zip(_.range(jids.length)).
        map(function(tuple){
            return deserialize(this._template(), this._locals(tuple));
        }, this).value();
    },
    _template: function() {
        return 'vcard_request';
    },
    _rosterJids: function() {
        if (!this._jids) 
            this._jids = _(this._stateData['roster'].children).chain().
            select(queryElement).
            pluck('children').
            flatten().
            map(function(c){ return c.attrs['jid']}).
            value();
        return this._jids;    
    },
    
    _locals: function(tuple) {
        return {jid: this._stateData['jid'], to: tuple[0], id: ('lb_vcard_get_' + (tuple[1] + 1))};
    },
    
    _idrule: /lb_vcard_get_(\d+)/,
    
    _vcardResponse: function(response) {
        return response.name === 'iq' && this._idrule.test(response.attrs['id']);
    },
    
    _vcardJid: function(response) {
        return this._rosterJids()[Number(this._idrule.exec(response.attrs['id'])[1])-1];
    },
    
    changeUsing: function(response) {
        if (!this._vcardResponse(response)) return;
        var vcardJid = this._vcardJid(response);
        this._stateData['vcards'] = this._stateData['vcards'] || {};
        if (!(vcardJid in this._stateData['vcards'])) {
            this._stateData['vcards'][vcardJid] = response;
            this._recievedVcards = (this._recievedVcards || 0) + 1;
        }
        return (this._recievedVcards < this._rosterJids().length) ? this : module.exports.Dummy;
    }
});

makeState('PreDomainItemsDisco', {
    _template: function() {
        return 'domain_items_disco';
    },
    _id: 'lb_domain_items_1',
    _locals: function() {
        return {jid: this._stateData['jid'], to: this._stateData['to'], id: this._id};
    },
    _isDomainItemsResponse: function(response){
        return response.name === 'iq' && response.attrs['id'] === this._id;   
    },
    _appendStateData: function(response) {
        this._stateData['domain_items'] = response;
    },
    _nextState: function(response) {
        if (this._isDomainItemsResponse(response))
            return 'PreServiceInfoDisco';
    }
});

makeState('PreServiceItemsDisco', {
    _template: function() {
        return 'service_items_disco';
    },
    _id: function() {
        return 'lb_service_items_' + this._serviceName;
    },
    _locals: function() {
        return {jid: this._stateData['jid'], serviceName: this._serviceName, id: this._id()};
    },
    
    _appendStateData: function(response) {
        this._stateData['rooms'] = this._stateData['rooms'] || {};
        this._stateData['rooms'][this._serviceName] = response;
    },
    
    _isServiceItemsResponse: function(response) {
        return response.name === 'iq' && response.attrs['id'] === this._id();   
    },  
    
    changeUsing: function(response) {
        if (this._isServiceItemsResponse(response)) {
            this._appendStateData(response);
            return new module.exports.RoomsFetched(this._stateData, this._serviceName);
        }
    }
}, function(stateData, serviceName){
    this._stateData = stateData || {};
    this._serviceName = serviceName;
})

makeState('PreServiceInfoDisco', {
    _createStanzas: function() {
        return _(this._services()).chain().
        zip(_.range(this._services().length)).
        reverse().
        map(function(serviceTuple){
            return deserialize(this._template(), this._locals(serviceTuple));
        }, this).value();
    },
    
    _services: function() {
        return _(this._stateData['domain_items'].children).find(queryElement).children;
    },
    
    _template: function() {
        return 'service_info_disco';
    },
    
    _locals: function(serviceTuple) {
        return {jid: this._stateData['jid'], to: serviceTuple[0].attrs['jid'], id: ('lb_service_disco_' + (serviceTuple[1] + 1))};
    },
    
    _isServiceInfoResponse: function(response) {
        return response.name === 'iq' && /lb_service_disco_\d+/.test(response.attrs['id']);
    },
    
    _isMucItem: function(response) {
        var query = _(response.children).find(queryElement);
        return query && _(query.children).any(function(child){
            return child.name === 'feature' && child.attrs['var'] === 'http://jabber.org/protocol/muc';
        });
    },
    
    _storeResponse: function(response) {
        this._responseIds = this._responseIds || {};
        if (!this._responseIds[response.attrs['id']]) {
            this._responseIds[response.attrs['id']] = response;
        }
    },
    
    _isDuplicateResponse: function(response) {
        return this._responseIds && !!this._responseIds[response.attrs['id']];
    },
    
    changeUsing: function(response) {
        if (this._isServiceInfoResponse(response)) {
            if (this._isDuplicateResponse(response)) return this;
            this._storeResponse(response);
            var nextState =  (_(this._responseIds).keys().length === this._services().length) ? 
                module.exports.Dummy : this;
            if (this._isMucItem(response)) {
                var newState = new module.exports.PreServiceItemsDisco(this._stateData, response.attrs['from']);
                return new module.exports.Composite([newState, nextState]);
            }
            return nextState;
        }
    }    
});

module.exports.Dummy = {
    done: function(){return true},
    responseStanzas: function(){return []},
    shouldRestartStream: function(){return false}
}

module.exports.Dummy.__defineGetter__('name', function(){
    return 'Dummy';
});

makeState('RosterFetched',{
    done: function() { 
        return true; 
    },
    
    _insertChild: function(node, newChild) {
        if (_(newChild).isString()) {
            node.t(newChild)
        } else {
            var elmnt = node.c(newChild.name, newChild.attrs);
            _(newChild.children).each(function(child){
                this._insertChild(elmnt, child);
                }, this);
        }
    },
    
    responseStanzas: function() {
        var roster = this._stateData['roster'];
        roster.attrs['id'] = 'lb_roster_1';
        roster.attrs['type'] = 'set';
        var query = _(roster.children).find(queryElement);
        query.attrs['xmlns'] = 'http://talk.to/x-roster';
        var vcardResponse = this._stateData['vcards'];
        query.children = _(query.children).map(function(item){
            if (vcardResponse && vcardResponse[item.attrs['jid']]) {
                this._insertChild(item, _(vcardResponse[item.attrs['jid']].children).find(function(c){
                    return c.name === 'vCard';
                }))
            }
            return item;
        }, this);
        return [roster];
    }
});

makeState('RoomsFetched', {
    done: function() { 
        return true; 
    },
    
    responseStanzas: function() {
        var rooms = this._stateData['rooms'][this._serviceName];
        rooms.attrs['id'] = 'lb_groups_1';
        rooms.attrs['type'] = 'set';
        var query = _(rooms.children).find(queryElement);
        query.attrs['xmlns'] = 'http://talk.to/x-groups';
        return [rooms];
    }
}, function(stateData, serviceName){
            this._stateData = stateData || {};
            this._serviceName = serviceName;
});

makeState('AuthFail', {
    done: function() {
        return true;
    },
    
    _template: function() {
        return 'auth_fail_response';
    },
    
    responseStanzas: function() {
        return [this._deserialize()];
    }
});

makeState('AuthSuccess', {
    done: function() {
        return true;
    },
    
    responseStanzas: function() {
        return [this._stateData['auth']]
    }
});

makeState('BindComplete', {
    done: function() {
        return true;
    },
    
    responseStanzas: function() {
        var bind = this._stateData['bind'];
        bind.attrs['type'] = 'set';
        return [bind];
    }
});

var Composite = module.exports.Composite = function(states){
    this._states = states;
};


Composite.prototype = {
    done: function() {
        return _(this._states).all(function(state){
            return state.done();
        });
    },

    popStanza: function() {
        return _(this._states).reduce(function(stanza, state){
            if (stanza) return stanza;
            if (!state.done()) return state.popStanza();
            }, null);
    },
    
    shouldRestartStream: function() {
        return _(this._states).any(function(state){
            return !state.done() && state.shouldRestartStream();
        });
    },
    
    changeUsing: function(stanza) {
        var changedStates = _(this._states).map(function(state){
            if (!state.done()) return state.changeUsing(stanza);
        });
        if (_(changedStates).all(_.isUndefined)) return;
        return new Composite(_(this._states).chain().
        zip(changedStates).
        map(function(stateSet){
            return stateSet[1] ? stateSet[1] : stateSet[0];
        }).
        value());
    },
    
    responseStanzas: function() {
        return _(this._states).chain().map(function(state){
            return state.responseStanzas();
        }).flatten().value();
    }
};

Composite.prototype.__defineGetter__("name", function(){
    return "Composite";
});

makeComposite = function(stateNames, stateData) {
    return new module.exports.Composite(_(stateNames).map(function(stateName){
        return new module.exports[stateName](stateData);
    }));
};  

