// -*- tab-width: 4 -*-

/* Directi Pvt. Ltd. 2011 */

/*
	logs following every minute[unit time say,on per unit time basis]
	
	(using a "quantized-window-concept" instead of "sliding-window-concept")
	
	1+. no. of sessions											[this.sessions.all.running]
	
	2+. no. of streams											[this.streams.all.running]
	
	3+. no. of sessions dropped									[this.prev_metric.inactive_sessions]
	
	4+. no. of requests processed								[this.prev_metric.requests_processed]
	
	5+. no. of empty responses sent								[this.prev_metric.empty_responses]
	
	6+. no. of streams by type of client						[this.devices['device_type'].running]
	
	7. no. of tcp connections by destination host				[todo(on the basis of to attrib already exists)]
	
	8+. no. of tcp connection failures while establishing		[this.prev_metric.tcp_conn_failures]
			
	9+. no. of tcp connection failures after establishment		[this.prev_metric.other_tcp_conn_failures]

points to ponder:

[definition: tcp connection failure ie. underlying xmpp stream is not closed but 'terminate' is fired]
[it will inform about no. of streams not sessions by type of client ...check it out]

*/


var util = require('util');
var us   = require('underscore');
var nxb  = require('node-xmpp-bosh');
var os   = require('os');

var stats = null;

// The maximum amount of time that zombie (terminated) 
// statistic objects are allowed to stick around.
var MAX_ZOMBIE_SEC = 3600;

// Run GC every GC_INTERVAL_SEC second
var GC_INTERVAL_SEC = 3600 * 5;

/*
metrics logger is refreshed every METRICS_INTERVAL_SEC second
*/
var METRICS_INTERVAL_SEC = 60;

function empty_stats_hash() {
	return {
		bytes_received:      0, 
		bytes_sent:          0, 
		bytes_transferred:   0, 
		packets_received:    0, 
		packets_sent:        0, 
		packets_transferred: 0
	};
}

function empty_unique_hash() {
    return {
        active: 0,
        total:  0
    }
}

function empty_aggregate_stats() {
	var ret = {
		running: 0,
		total: 0,
		total_time: 0
	};

	ret.__defineGetter__('average_time', function() {
		var den = this.total - this.running;
		den = den > 0 ? den : 1;
		return nxb.dutil.time_diff(new Date(0), new Date(this.total_time / den));
	});

	return ret;
}

/*
empty metrics box provider
*/
function empty_metrics() {
	return {
		timestamp 				: new Date(),
		inactive_sessions		: 0,
		requests_processed		: 0,
		empty_responses			: 0,
		tcp_conn_failures		: 0,
		other_tcp_conn_failures	: 0,
		reInit					: function() {
									this.timestamp 					= new Date();
									this.inactive_sessions			= 0;
									this.requests_processed			= 0;
									this.empty_responses			= 0;
									this.tcp_conn_failures			= 0;
									this.other_tcp_conn_failures	= 0;
							  	  }
	};
}


function get_device_type (sstate) {
    var ua = sstate.state.ua;
    return ua == 'android' || ua == 'iphone' || ua == 'desktop' || ua == 'blackberry'?ua:'unknown';
}

function Stats(server, webrepl) {
	this.server   = server;
	this.webrepl  = webrepl;

	this.all      = empty_stats_hash();
	this.streams  = { };
	this.sessions = { };

    this.jids     = { };
    this.domains  = { };
    this.devices  = { };
    
	/*
	get empty objects to keep things for the previous window and current window
	*/
	this.curr_metric = empty_metrics();
	this.prev_metric = empty_metrics();
	
    this.jids.all     = empty_unique_hash();
    this.domains.all  = empty_unique_hash();
	this.streams.all  = empty_aggregate_stats();
	this.sessions.all = empty_aggregate_stats();
    
    this.cpu = { };
    this.cpu.__defineGetter__('uptime', function() {
        return os.uptime();
    });
    this.cpu.__defineGetter__('loadavg', function() {
        return os.loadavg();
    });
    this.cpu.__defineGetter__('totalmem', function() {
        return os.totalmem();
    });
    this.cpu.__defineGetter__('freemem', function() {
        return os.freemem();
    });
    this.cpu.__defineGetter__('cpus', function() {
        return os.cpus();
    });

    this.process = { };
    this.process.__defineGetter__('pid', function() {
        return process.pid;
    });
    this.process.__defineGetter__('memory_usage', function() {
        return process.memoryUsage();
    });
    
	this.init();
}

function Temp() { }
Temp.prototype = {
	_update_sent: function(hash, packet, packet_len) {
		hash.bytes_sent          += packet_len;
		hash.bytes_transferred   += packet_len;
		hash.packets_sent        += 1;
		hash.packets_transferred += 1;
	}, 

	_update_received: function(hash, packet, packet_len) {
		hash.bytes_received      += packet_len;
		hash.bytes_transferred   += packet_len;
		hash.packets_received    += 1;
		hash.packets_transferred += 1;
	}, 

	//Received by BOSH (from the XMPP server)
	packet_received: function(packet, sstate) {
		var stream  = this.streams[sstate.name];
		var session = this.sessions[sstate.state.sid];
        var jid     = sstate.from?this.jids[sstate.from]:null;
        var domain  = this.domains[sstate.to];
        var device  = this.devices[get_device_type(sstate)];
		var packet_len = packet.toString().length;

		this._update_received(this.all, packet, packet_len);

		if (stream) {
			this._update_received(stream, packet, packet_len);
		}

		if (session) {
			this._update_received(session, packet, packet_len);
		}
        
        if (jid) {
			this._update_received(jid, packet, packet_len);
		}

        if (domain) {
            this._update_received(domain, packet, packet_len);
        }
        
        if (device) {
            this._update_received(device, packet, packet_len);
        }
	}, 

	//Sent by BOSH (to the XMPP server)
	packet_sent: function(packet, sstate) {
		var stream  = this.streams[sstate.name];
		var session = this.sessions[sstate.state.sid];
        var jid     = sstate.from?this.jids[sstate.from]:null;
        var domain  = this.domains[sstate.to];
        var device  = this.devices[get_device_type(sstate)];
		var packet_len = packet.toString().length;

		// console.log(stream, session, packet_len);

		this._update_sent(this.all, packet, packet_len);

		if (stream) {
			this._update_sent(stream, packet, packet_len);
		}

		if (session) {
			this._update_sent(session, packet, packet_len);
		}
        
        if (jid) {
            this._update_sent(jid, packet, packet_len);
        }
        
        if (domain) {
            this._update_sent(domain, packet, packet_len);
        }

        if (device) {
            this._update_sent(device, packet, packet_len);
        }
    }, 

	packets_sent: function(packets, sstate) {
		packets.forEach(us.bind(function(packet) {
			this.packet_sent(packet, sstate);
		}, this));
	}, 

	stream_added: function(sstate) {
        function increment_counters(hash) {
            hash.running += 1;
            hash.total   += 1;
        }
 
		var stream = empty_stats_hash();
		stream.start_time  = new Date();
		stream.end_time    = null;
        stream.device_type = get_device_type(sstate);
        stream.from        = sstate.from?sstate.from:null;
       
		// console.log(sstate.name);

		this.streams[sstate.name] = stream;
		increment_counters(this.streams.all);
        
		var session = empty_stats_hash();
		if (!this.sessions[sstate.state.sid]) {
			this.sessions[sstate.state.sid] = session;
			session.start_time = new Date();
			session.end_time   = null;
			increment_counters(this.sessions.all);
		}
        
        function _make_object() {
            var obj     = empty_stats_hash();
            obj.streams             = { };
            obj.streams.active      = { };
            obj.streams.terminated  = { };
            us.extend(obj, empty_aggregate_stats());
            return obj;
        }

        function _conf_object(obj) {
            obj.streams.active[sstate.name] = stream;
            increment_counters(obj);
        }

        var to = sstate.to;
        if (to) {
            if (!this.domains[to]) {
                this.domains[to] = _make_object();
                this.domains.all.total  += 1;
            }
            if(this.domains[to].running === 0) {
                this.domains.all.active += 1;
            }
            _conf_object(this.domains[to]);
        }
        
        var device_type = get_device_type(sstate);
        if (device_type) {
            if (!this.devices[device_type]) {
                this.devices[device_type] = _make_object();
            }
            _conf_object(this.devices[device_type]);
        }
        
        var from = sstate.from;
        if (from) {
            if (!this.jids[from]) {
                this.jids[from]                  = _make_object();
                this.jids[from].last_created     = { };
                this.jids[from].last_terminated  = { };
                this.jids.all.total       += 1;
            }
            if(this.jids[from].running === 0){
                this.jids.all.active += 1;
            }
            _conf_object(this.jids[from]);
            this.jids[from].last_created.stream  = sstate.name;
            this.jids[from].last_created.time    = stream.start_time;
        }
	}, 
	
	/*
	distinguishing between stream termination from server and by client
	*/
	terminated_by_client: function(sstate) {
		this.stream_terminated(sstate, false);
	},
	
	terminated_by_server: function(sstate) {
		this.stream_terminated(sstate, true);
	},
	
	stream_terminated: function(sstate, is_server) {
        var stream = this.streams[sstate.name];
        
        /*
		checking for tcp connection failure
		stream_terminate is the event listener for both 'terminate' [is_server = true] and 'stream-terminate'
       	detect here the exact scenario ie. when xmpp stream on this is not closed yet but terminate is fired
		*/
        if(is_server === true && sstate.can_be_tcp_conn_failure === true)
        {
        	if(stream)
        	{
        		//error after the tcp-connection establishment as stream already existed
        		this.curr_metric.other_tcp_conn_failures++;
        	}
        	else
        	{
        		//error at the time of establishing tcp-connection
        		this.curr_metric.tcp_conn_failures++;
			}
        }
        
        if (!stream || stream.end_time) {
            return;
        }

        function _update_stats(hash) {
            var hash_active_stream = hash?hash.streams.active[sstate.name]:null;
            if (hash_active_stream) {
                hash.running     -= 1;
                hash.total_time  += (new Date() - hash_active_stream.start_time);
                hash.streams.terminated[sstate.name] = hash_active_stream;
                delete hash.streams.active[sstate.name];
            }
        }

	    var to = sstate.to;
        if (to) {
            _update_stats(this.domains[to]);
            if (this.domains[to].running === 0) {
                this.domains.all.active -= 1;
            }
        }

        var device_type = get_device_type(sstate);
        if (device_type) {
            _update_stats(this.devices[device_type]);
        }
        
        var from = sstate.from;
        if (from) {
            var jid_active_stream = this.jids[from]?this.jids[from].streams.active[sstate.name]:null;
            if (jid_active_stream) {
                this.jids[from].last_terminated.stream  = sstate.name;
                this.jids[from].last_terminated.time    = jid_active_stream.end_time;
            }
            _update_stats(this.jids[from]);
            if (this.jids[from].running === 0) {
                this.jids.all.active -= 1;
            }
        }
        

		if (stream && !stream.end_time) {
			stream.end_time = new Date();
            stream.terminate_condition   = sstate.condition;
			this.streams.all.running    -= 1;
			this.streams.all.total_time += (stream.end_time - stream.start_time);
		}

		var session = this.sessions[sstate.state.sid];
		if (session && sstate.state.streams.length === 0 && !session.end_time) {
			session.end_time = new Date();
			this.sessions.all.running    -= 1;
			this.sessions.all.total_time += (session.end_time - session.start_time);
			
			/*
			last stream of the session and hence session is being destroyed
			we have to keep track of the no. of sessions dropped in last one minute
			so update our special counters.
			*/
			this.curr_metric.inactive_sessions++;
		}
	}, 
	
	/*
	special event['empty-response-sent'] listener
	*/
	empty_response_sent: function(body) {
		/*
			This event(empty-response-sent) is emitted when a 'held http-response object' is replied back empty
			after wait-Timeout timer is triggred.
		*/
		this.curr_metric.empty_responses++;
	},
	
	/*
	special event ['request-came'] listener
	*/
	request_came: function(body) {
		/*
		This event(request-came) is emitted when a 'http-post-request' 
		comes to bosh (in particular when process_bosh_request() is invoked)
		we are to keeptrack of number of requests entertained by bosh  
		*/
		this.curr_metric.requests_processed++;	
	},
	
	run_gc: function() {
		function truncate_zombies(hash) {
			var now = new Date();
			var hash_keys = Object.keys(hash).filter(function(key) {
				return key != 'all';
			});

			hash_keys.forEach(function(key) {
				var obj = hash[key];
				if (obj.end_time && (now - obj.start_time > MAX_ZOMBIE_SEC * 1000)) {
					delete obj[key];
				}
			});
		}

		truncate_zombies(this.sessions);
		truncate_zombies(this.streams);
	},
	
	/*
	refresh merics
	reinitialises every thing to log 
	*/
	refresh_metrics: function() {
				
		this.prev_metric.reInit();
		
		// swap current and previous
		var temp =  this.prev_metric;
		this.prev_metric = this.curr_metric;
		this.curr_metric = temp;
		
		//log to the console [comment it out after testing]
		console.log("METRIC_STATS: ");
		for(property in this.prev_metric)
		{
			if(this.prev_metric.hasOwnProperty(property) && typeof(this.prev_metric[property]) !== 'function')
			{
				console.log(property + " : " + this.prev_metric[property]);
			}
		}
		console.log("numberOfSessions :" + this.sessions.all.running);
		console.log("numberOfStreams :" + this.streams.all.running);
		/*
		if(this.devices['android'].running)
			console.log("devices['android'] :" + this.devices['android'].running);
		if(this.devices['iphone'].running)	
			console.log("devices['iphone'] :" + this.devices['iphone'].running);
		if(this.devices['desktop'].running)	
			console.log("devices['desktop'] :" + this.devices['desktop'].running);
		if(this.devices['blackberry'].running)	
			console.log("devices['blackberry'] :" + this.devices['blackberry'].running);
		*/
	},
	
	init: function() {
		this.packet_received   = us.bind(this.packet_received, this);
		this.stream_added      = us.bind(this.stream_added, this);
		this.packets_sent      = us.bind(this.packets_sent, this);
		
		/*
		distinguish between 2 stream termination 
		*/
		this.terminated_by_client = us.bind(this.terminated_by_client, this);
		this.terminated_by_server = us.bind(this.terminated_by_server, this);
		
		/*
		adding 2 special events 
		*/
		this.empty_response_sent	= us.bind(this.empty_response_sent, this);
		this.request_came			= us.bind(this.request_came, this);
	
		
		this.server.on('stream-added', this.stream_added);
		this.server.on('stream-terminate', this.terminated_by_client);
		this.server.on('terminate', this.terminated_by_server);
		this.server.on('nodes', this.packets_sent);
		this.server.on('response', this.packet_received);
		
		/*
		listening for 2 special events
		*/
		this.server.on('empty-response-sent', this.empty_response_sent);
		this.server.on('request-received', this.request_came);
		
		
		this.gc_interval = setInterval(us.bind(this.run_gc, this), GC_INTERVAL_SEC * 1000);
		
		/*
		set Interval timer for metrics refresh
		*/
		this.metrics_interval = setInterval(us.bind(this.refresh_metrics, this), METRICS_INTERVAL_SEC * 1000);
		
	}, 

	destroy: function() {
		this.server.removeListener('stream-added', this.stream_added);
		this.server.removeListener('stream-terminate', this.terminated_by_client);
		this.server.removeListener('terminate', this.terminated_by_server);
		this.server.removeListener('nodes', this.packets_sent);
		this.server.removeListener('response', this.packet_received);
		
		clearInterval(this.gc_interval);
		
		/*
		clearing metric refresher 
		*/
		clearInterval(this.metrics_interval);

	}
};

// We need this hack to make webrepl autocomplete our members for us
util.inherits(Stats, Temp);

exports.init = function(ctor_args) {
    if (!ctor_args || !ctor_args.server || !ctor_args.webrepl) {
		throw new Error("You MUST set 'server', and 'webrepl' to their respective values");
    }

	console.log("INIT STATS MODULE");

    stats = new Stats(ctor_args.server, ctor_args.webrepl);
	ctor_args.webrepl.context.stats = stats;
};

exports.destroy = function() {
	if (stats) {
		stats.destroy();
		delete stats.webrepl.context.stats;
		stats = null;
	}
};
