require(__dirname + '/../test_utils');
var assert = require('assert');
var _ = require('underscore'); 
var BoosterFilter = require(__dirname + '/../../plugins/login_booster/booster_filter');
var BoosterClient = require(__dirname + '/../../plugins/login_booster/booster_client').BoosterClient;

batman_account = {name: "batman@facebook.com"};
bruce_wayne_account = {name: "brucewayne@iamrich.com"}

function test_with_gently(test) {
    return function() {
        var gently = new (require('gently'));
        test(gently);
        gently.verify();
    }
};

assert.suppresses_event = function(value){
    return assert.equal(false, value, "Event has not been suppressed.");
};

assert.does_not_suppress_event = function(value){
    return assert.isUndefined(value, "Event has been suppressed.");
};

var createBoosterFilter = function(server) {
    if (!server) {
        server = {};
        server.emit = function() {};
    }
    return BoosterFilter(server);
};

var events = require("events");

BoosterClient.create = function() { 
    return _(new events.EventEmitter()).tap(function(client){
        client.consumes = function(){};
    })
};

var test_with_mock_booster_client = function(test) {
    return test_with_gently(function(gently){
        var client = BoosterClient.create();
        gently.expect(BoosterClient, "create", function() {
            return client;
        });
        gently.expectResponse = function(server, data, sstate) {
            gently.expect(server, 'emit', function(event, stanza, actualSstate){
                assert.equal('response', event);
                assert.equal(data, stanza);
                assert.equal(sstate, actualSstate);
            });
        };
        gently.expectNoResponse = function(server, data, sstate) {};
        test(gently, client);
    });
};

var return_true_for_actual_roster = function(stanza) {
    return stanza.attrs['xmlns'] === 'jabber:iq:roster';
}

module.exports = {
    'does not supress auth packets which are not talk.to extensions': function(){
        var booster = createBoosterFilter();
        assert.does_not_suppress_event(booster.nodes(_.createNodes('legacy_auth'),batman_account));
    },
    
    'suppresses auth packets which are talk.to extensions': function() {
        var booster = createBoosterFilter();
        assert.suppresses_event(booster.nodes(_.createNodes('talkto_auth'), batman_account));
    },
    
    'does not suppress responses by default': function() {
        var booster = createBoosterFilter();
        assert.does_not_suppress_event(booster.response(_.createStanza('auth_stream_features'), batman_account));
    },
    
    'does not suppress responses after legacy auth packet': function() {
        var booster = createBoosterFilter();
        booster.nodes(_.createNodes('legacy_auth'), {name: "batman@facebook.com"});
        assert.does_not_suppress_event(booster.response(_.createStanza('auth_stream_features'), batman_account));        
    }, 
    
    'suppresses responses after talk.to auth packet': function() {
        var booster = createBoosterFilter();
        booster.nodes(_.createNodes('talkto_auth'), batman_account);
        assert.suppresses_event(booster.response(_.createStanza('auth_stream_features'), batman_account));        
    },
    
    'does not suppress responses of other streams when a talk.to auth packet comes on one stream': function() {
        var booster = createBoosterFilter();
        booster.nodes(_.createNodes('talkto_auth'), batman_account);
        assert.does_not_suppress_event(booster.response(_.createStanza('auth_stream_features'), bruce_wayne_account));        
    },
    
    'creates a BoosterClient when a talk.to auth packet comes': test_with_gently(function(gently){
        var booster = createBoosterFilter();
        var talkto_auth_stanzas = _.createNodes('talkto_auth');
        gently.expect(BoosterClient, "create", function(authStanzas) {
            assert.equal(talkto_auth_stanzas, authStanzas);
            return BoosterClient.create();
        });
        booster.nodes(talkto_auth_stanzas, batman_account);
    }),
    
    'passes every response packet to the BoosterClient to consume': test_with_mock_booster_client(function(gently, client) {
        var booster = createBoosterFilter();
        booster.nodes(_.createNodes('talkto_auth'), batman_account);
        var dummy = _.createStanza('dummy');
        gently.expect(client, "consumes", function(stanza){
            assert.equal(dummy, stanza)
        });
        booster.response(dummy, batman_account);
    }),
    
    'stops suppressing responses when a BoosterClient publishes a done': test_with_mock_booster_client(function(gently, client) {
        var booster = createBoosterFilter();
        booster.nodes(_.createNodes('talkto_auth'), batman_account);
        client.emit('done', []);
        assert.does_not_suppress_event(booster.response(_.createStanza('dummy'), batman_account));
    }),
    
    'emits response events containing the packets from the BoosterClient done event': test_with_mock_booster_client(function(gently, client) {
        var server = {};
        var booster = createBoosterFilter(server);
        booster.nodes(_.createNodes('talkto_auth'), batman_account);
        var roster = _.createStanza('x-roster'), 
            groups = _.createStanza('x-groups');
        gently.expectResponse(server, roster, batman_account);
        gently.expectResponse(server, groups, batman_account);
        client.emit('done', [roster, groups]);
    }),
    
    'emits response events containing all packets that the BoosterClient did not consume after all the packets emitted by the BoosterClient when it publishes a done': test_with_mock_booster_client(function(gently, client) {
        var server = {};
        var booster = createBoosterFilter(server);
        booster.nodes(_.createNodes('talkto_auth'), batman_account);
        var roster_from_booster = _.createStanza('x-roster'), 
            groups_from_booster = _.createStanza('x-groups');
        var message_from_server = _.createStanza('message'),
            roster_from_server = _.createStanza('actual_roster');
        var stubFn = function(st) {
            if (st === message_from_server) return false;
            if (st === roster_from_server) return true;
        }    
        gently.expect(client, 'consumes', 2, stubFn);    
        
        gently.expectResponse(server, roster_from_booster, batman_account);
        gently.expectResponse(server, groups_from_booster, batman_account);        
        gently.expectResponse(server, message_from_server, batman_account);
        gently.expectNoResponse(server, roster_from_server, batman_account); // this line is the point of this entire test
    
        booster.response(message_from_server, batman_account);
        booster.response(roster_from_server, batman_account);
        client.emit('done', [roster_from_booster, groups_from_booster]);
    }),
    
    'Restarts the xmpp stream when it recieves a stream-restart event': test_with_mock_booster_client(function(gently, client) {
        var server = {};
        var booster = createBoosterFilter(server);
        booster.nodes(_.createNodes('talkto_auth'), batman_account);
        gently.expect(server, 'emit', function(event, sstate){
            assert.equal('stream-restart', event);
            assert.equal(batman_account, sstate);
        });
        client.emit('stream-restart');
    }),
    
    'BoosterClient receives a function with which it can publish a nodes event': test_with_gently(function(gently) {
        var server = {};
        var booster = createBoosterFilter(server);
        var publish;
        gently.expect(BoosterClient, 'create', function(talkto_auth_stanzas, passed_publish){
            publish = passed_publish;
            return BoosterClient.create();
        });
        gently.expect(server, 'emit', function(event, nodes, sstate){
            assert.equal('nodes', event);
            assert.length(nodes, 1);
            assert.equal(legacy_auth, nodes[0]);
            assert.equal(batman_account, sstate);
        });
        booster.nodes(_.createNodes('talkto_auth'), batman_account);
        var legacy_auth = _.createNodes('legacy_auth');
        publish(legacy_auth);
    }),
    
    'it does something if a second talk.to auth packet comes when there is already an talk.to auth packet': _.to_be_implemented,
    'it does something to prevent the restart handler from leaking memory incase restart is not fired': _.to_be_implemented,
	'it does something if the connection closes with the server': _.to_be_implemented

};

